# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, division

import numpy as np

from PIL import Image, ImageDraw, ImageFont
import cv2
import random

from libs.label_name_dict.label_dict import LabelMap
from utils.tools import get_dota_short_names
from libs.utils.coordinate_convert import forward_convert
from dataloader.pretrained_weights.pretrain_zoo import PretrainModelZoo


class DrawBox(object):

    def __init__(self, cfgs):
        self.cfgs = cfgs
        label_map = LabelMap(cfgs)
        self.name_label_map, self.label_name_map = label_map.name2label(), label_map.label2name()

        self.NOT_DRAW_BOXES = 0
        self.ONLY_DRAW_BOXES = -1
        self.ONLY_DRAW_BOXES_WITH_SCORES = -2

        self.STANDARD_COLORS = [
            'AliceBlue', 'Chartreuse', 'Aqua', 'Aquamarine', 'Azure', 'Beige', 'Bisque',
            'BlanchedAlmond', 'BlueViolet', 'BurlyWood', 'CadetBlue', 'AntiqueWhite',
            'Chocolate', 'Coral', 'CornflowerBlue', 'Cornsilk', 'Crimson', 'Cyan',
            'DarkCyan', 'DarkGoldenRod', 'DarkGrey', 'DarkKhaki', 'DarkOrange',
            'DarkOrchid', 'DarkSalmon', 'DarkSeaGreen', 'DarkTurquoise', 'DarkViolet',
            'DeepPink', 'DeepSkyBlue', 'DodgerBlue', 'FireBrick', 'FloralWhite',
            'ForestGreen', 'Fuchsia', 'Gainsboro', 'GhostWhite', 'Gold', 'GoldenRod',
            'Salmon', 'Tan', 'HoneyDew', 'HotPink', 'IndianRed', 'Ivory', 'Khaki',
            'Lavender', 'LavenderBlush', 'LawnGreen', 'LemonChiffon', 'LightBlue',
            'LightCoral', 'LightCyan', 'LightGoldenRodYellow', 'LightGray', 'LightGrey',
            'LightGreen', 'LightPink', 'LightSalmon', 'LightSeaGreen', 'LightSkyBlue',
            'LightSlateGray', 'LightSlateGrey', 'LightSteelBlue', 'LightYellow', 'Lime',
            'LimeGreen', 'Linen', 'Magenta', 'MediumAquaMarine', 'MediumOrchid',
            'MediumPurple', 'MediumSeaGreen', 'MediumSlateBlue', 'MediumSpringGreen',
            'MediumTurquoise', 'MediumVioletRed', 'MintCream', 'MistyRose', 'Moccasin',
            'NavajoWhite', 'OldLace', 'Olive', 'OliveDrab', 'Orange', 'OrangeRed',
            'Orchid', 'PaleGoldenRod', 'PaleGreen', 'PaleTurquoise', 'PaleVioletRed',
            'PapayaWhip', 'PeachPuff', 'Peru', 'Pink', 'Plum', 'PowderBlue', 'Purple',
            'Red', 'RosyBrown', 'RoyalBlue', 'SaddleBrown', 'Green', 'SandyBrown',
            'SeaGreen', 'SeaShell', 'Sienna', 'Silver', 'SkyBlue', 'SlateBlue',
            'SlateGray', 'SlateGrey', 'Snow', 'SpringGreen', 'SteelBlue', 'GreenYellow',
            'Teal', 'Thistle', 'Tomato', 'Turquoise', 'Violet', 'Wheat', 'White',
            'WhiteSmoke', 'Yellow', 'YellowGreen', 'LightBlue', 'LightGreen'
        ]
        self.FONT = ImageFont.load_default()

    def find_head_edge(self, box, head):
        head_dict = {0: '11', 1: '10', 2: '00', 3: '01'}
        flag = head_dict[int(head)]
        box[4] += random.random() * 0.1
        box_eight = forward_convert(np.array([box]), False)[0]
        box_eight = np.reshape(box_eight, [4, 2])
        four_edges = [[box_eight[0], box_eight[1]], [box_eight[1], box_eight[2]],
                      [box_eight[2], box_eight[3]], [box_eight[3], box_eight[0]]]
        for i in range(4):
            center_x = (four_edges[i][0][0] + four_edges[i][1][0]) / 2.
            center_y = (four_edges[i][0][1] + four_edges[i][1][1]) / 2.
            if (center_x - box[0]) >= 0 and (center_y - box[1]) >= 0:
                res = '11'
                if res == flag:
                    return four_edges[i]
            elif (center_x - box[0]) >= 0 and (center_y - box[1]) <= 0:
                res = '10'
                if res == flag:
                    return four_edges[i]
            elif (center_x - box[0]) <= 0 and (center_y - box[1]) <= 0:
                res = '00'
                if res == flag:
                    return four_edges[i]
            else:
                res = '01'
                if res == flag:
                    return four_edges[i]

    def draw_head(self, draw_obj, box, head, color, width=3):
        head_edge = self.find_head_edge(box, head)
        if head_edge is None:
            pass
        else:
            draw_obj.line(xy=[(head_edge[0][0], head_edge[0][1]), (head_edge[1][0], head_edge[1][1])],
                          fill='Red',
                          width=width+1)
            center_x = (head_edge[0][0] + head_edge[1][0]) / 2.
            center_y = (head_edge[0][1] + head_edge[1][1]) / 2.
            draw_obj.line(xy=[(center_x, center_y), (box[0], box[1])],
                          fill=color,
                          width=width)

    def draw_a_rectangel_in_img(self, draw_obj, box, color, width, method):
        '''
        use draw lines to draw rectangle. since the draw_rectangle func can not modify the width of rectangle
        :param draw_obj:
        :param box: [x1, y1, x2, y2]
        :return:
        '''
        # color = (0, 255, 0)
        if method == 0:
            x1, y1, x2, y2 = box[0], box[1], box[2], box[3]
            top_left, top_right = (x1, y1), (x2, y1)
            bottom_left, bottom_right = (x1, y2), (x2, y2)

            draw_obj.line(xy=[top_left, top_right],
                          fill=color,
                          width=width)
            draw_obj.line(xy=[top_left, bottom_left],
                          fill=color,
                          width=width)
            draw_obj.line(xy=[bottom_left, bottom_right],
                          fill=color,
                          width=width)
            draw_obj.line(xy=[top_right, bottom_right],
                          fill=color,
                          width=width)
        elif method == 1:
            x_c, y_c, w, h, theta = box[0], box[1], box[2], box[3], box[4]
            rect = ((x_c, y_c), (w, h), theta)
            rect = cv2.boxPoints(rect)
            rect = np.int0(rect)
            draw_obj.line(xy=[(rect[0][0], rect[0][1]), (rect[1][0], rect[1][1])],
                          fill=color,
                          width=width)
            draw_obj.line(xy=[(rect[1][0], rect[1][1]), (rect[2][0], rect[2][1])],
                          fill=color,
                          width=width)
            draw_obj.line(xy=[(rect[2][0], rect[2][1]), (rect[3][0], rect[3][1])],
                          fill=color,
                          width=width)
            draw_obj.line(xy=[(rect[3][0], rect[3][1]), (rect[0][0], rect[0][1])],
                          fill=color,
                          width=width)
        elif method == 2:
            for ii in range(box.shape[0] // 2):
                draw_obj.line(xy=[(box[(2 * ii) % box.shape[0]], box[(2 * ii + 1) % box.shape[0]]),
                                  (box[(2 * ii + 2) % box.shape[0]], box[(2 * ii + 3) % box.shape[0]])],
                              fill=color,
                              width=width)
        else:
            pass

    def only_draw_scores(self, draw_obj, box, score, color):

        x, y = box[0], box[1]
        draw_obj.rectangle(xy=[x, y, x+60, y+10],
                           fill=color)
        draw_obj.text(xy=(x, y),
                      text="obj:" + str(round(score, 2)),
                      fill='black',
                      font=self.FONT)

    def draw_label_with_scores(self, draw_obj, box, label, score, color):
        x, y = box[0], box[1]
        draw_obj.rectangle(xy=[x, y, x + 60, y + 10],
                           fill=color)

        txt = self.label_name_map[label] + ':' + str(round(score, 2))
        draw_obj.text(xy=(x, y),
                      text=txt,
                      fill='black',
                      font=self.FONT)

    def draw_label_with_scores_csl(self, draw_obj, box, label, score, method, head, color='White'):
        x, y = box[0], box[1]
        draw_obj.rectangle(xy=[x, y, x + 60, y + 10],
                           fill='White')

        if self.cfgs.DATASET_NAME.startswith('DOTA'):
            label_name = get_dota_short_names(self.label_name_map[label])
        else:
            label_name = self.label_name_map[label]
        txt = label_name + ':' + str(round(score, 2))
        # txt = ' ' + label_name
        draw_obj.text(xy=(x, y),
                      text=txt,
                      fill='black',
                      font=self.FONT)
        if method == 1:
            draw_obj.rectangle(xy=[x, y + 10, x + 60, y + 20],
                               fill='White')
            if self.cfgs.ANGLE_RANGE == 180:
                if box[2] < box[3]:
                    angle = box[-1] + 90
                else:
                    angle = box[-1]
            else:
                angle = box[-1]
            txt_angle = 'angle:%.1f' % angle
            # txt_angle = ' %.1f' % angle
            draw_obj.text(xy=(x, y + 10),
                          text=txt_angle,
                          fill='black',
                          font=self.FONT)
            if head != -1:
                draw_obj.rectangle(xy=[x, y + 20, x + 60, y + 30],
                                   fill='White')
                txt_head = 'head:%d' % head
                draw_obj.text(xy=(x, y + 20),
                              text=txt_head,
                              fill='black',
                              font=self.FONT)
                self.draw_head(draw_obj, box, head, color)

    def draw_boxes_with_label_and_scores(self, img_array, boxes, labels, scores, method, head=None, is_csl=False, in_graph=True):
        if in_graph:
            pretrain_zoo = PretrainModelZoo()
            if self.cfgs.NET_NAME in pretrain_zoo.pth_zoo or self.cfgs.NET_NAME in pretrain_zoo.mxnet_zoo:
                img_array = (img_array * np.array(self.cfgs.PIXEL_STD) + np.array(self.cfgs.PIXEL_MEAN_)) * 255
            else:
                img_array = img_array + np.array(self.cfgs.PIXEL_MEAN)
        if method == 3:
            img_array = self.draw_contours(img_array, boxes, labels)
        elif method == 4:
            img_array = self.draw_boxes_ellipse(img_array, boxes, labels)

        img_array.astype(np.float32)
        boxes = boxes.astype(np.float32)
        labels = labels.astype(np.int32)
        img_array = np.array(img_array * 255 / np.max(img_array), dtype=np.uint8)

        img_obj = Image.fromarray(img_array)
        raw_img_obj = img_obj.copy()

        draw_obj = ImageDraw.Draw(img_obj)
        num_of_objs = 0

        if head is None:
            head = np.ones_like(labels) * -1

        for box, a_label, a_score, a_head in zip(boxes, labels, scores, head):

            if a_label != self.NOT_DRAW_BOXES:
                num_of_objs += 1
                self.draw_a_rectangel_in_img(draw_obj, box, color=self.STANDARD_COLORS[a_label], width=3, method=method)
                if a_label == self.ONLY_DRAW_BOXES:  # -1
                    continue
                elif a_label == self.ONLY_DRAW_BOXES_WITH_SCORES:  # -2
                     self.only_draw_scores(draw_obj, box, a_score, color='White')
                else:
                    if is_csl:
                        self.draw_label_with_scores_csl(draw_obj, box, a_label, a_score, method, a_head, color='White')
                    else:
                        self.draw_label_with_scores(draw_obj, box, a_label, a_score, color='White')

        out_img_obj = Image.blend(raw_img_obj, img_obj, alpha=0.7)

        return np.array(out_img_obj)

    def draw_boxes_ellipse(self, img_array, boxes, labels):
        labels = labels.astype(np.int32)
        for box, a_label in zip(boxes, labels):
            if box[2] < box[3]:
                box[4] += 90
            img_array = cv2.ellipse(img_array, center=(box[0], box[1]),
                                    axes=(max(box[2], box[3]), min(box[2], box[3])),
                                    angle=box[4], startAngle=0, endAngle=360,
                                    color=(255, 0, 0),
                                    thickness=3, lineType=-1)
        return img_array

    def draw_contours(self, img_array, polys, labels):
        for poly, a_label in zip(polys, labels):
            img_array = cv2.fillConvexPoly(img_array, np.reshape(poly, [-1, 2]).astype(np.int32), color=(255, 255, 255))
        return img_array











import math
import random
import numpy as np
import matplotlib.pyplot as plt
import alphashape

from libs.utils.mask_sample import points_sampling


def carttopolar(x, y, x0=0, y0=0):
    '''
    cartisian to polar coordinate system with origin shift to x0,y0
    '''
    x1 = x - x0
    y1 = y - y0
    # print('(x0,y0)sort',x0,y0)
    r = np.sqrt(x1 ** 2 + y1 ** 2)
    t = np.arctan2(y1, x1) * 180 / math.pi
    if y1 < 0:
        t += 360
    # print('x,y,r,t',x,y,r,t)
    return r, t


def sort_aniclkwise(xy_list,x0=None,y0=None):
    '''
    Sort points anit clockwise with x0 y0 as origin
    '''
    if x0 is None and y0 is None:
        (x0, y0) = np.mean(xy_list, axis=0).tolist()
    elif x0 is None:
        (x0, _) = np.mean(xy_list, axis=0).tolist()
    elif y0 is None:
        (_, y0) = np.mean(xy_list, axis=0).tolist()
    # print('origin used:',[x0,y0])

    for i in range(len(xy_list)):
          xy_list[i].append(i)

    xy_list1 = sorted(xy_list, key=lambda a_entry: carttopolar(a_entry[0], a_entry[1],x0,y0)[1])

    sort_index = []  
    for x in xy_list1:
        sort_index.append(x[2])
        del x[2]

    return np.array(xy_list1, np.float32)


def sort_aniclkwise_batch(polys):
    return np.array([sort_aniclkwise(poly.tolist()) for poly in polys], np.float32)


def draw(pts):
    color_dict = ['b', 'g', 'r', 'c', 'm', 'y', 'k']
    idx = random.randint(0,7)
    pts = np.array(pts).reshape(-1,2)
    xs = pts[:, 0]
    ys = pts[:, 1]
    plt.scatter(xs,ys)
    plt.plot(xs, ys)


def concave_hull(points, point_num, alpha=0.95):
    points = np.array(points)
    alpha *= alphashape.optimizealpha(points)
    hull = alphashape.alphashape(points, alpha)
    hull_pts = hull.exterior.coords.xy
    hull_pts = np.array(hull_pts)
    hull_pts = np.concatenate([np.reshape(hull_pts[0][:-1], [-1, 1]), np.reshape(hull_pts[1][:-1], [-1, 1])], axis=1)
    hull_pts = np.reshape(hull_pts, [-1, 2])
    hull_pts = points_sampling(hull_pts, point_num)
    return hull_pts


def concave_hull_batch(polys, point_num, alpha):
    return np.array([concave_hull(poly, point_num, alpha) for poly in polys], np.float32)


if __name__ == '__main__':
    pts = [[1, 1], [0, 2], [2, 1], [2, 2], [2, 0], [0, 1], [1, 0], [0, 0]]
    points = np.array([(17, 158), (15, 135), (38, 183), (43, 19), (93, 88), (96, 140), (149, 163), (128, 248), (216, 265),
              (248, 210), (223, 167), (256, 151), (331, 214), (340, 187), (316, 53), (298, 35), (182, 0), (121, 42)])

    # draw(points)
    # spts = sort_aniclkwise(pts)
    spts = concave_hull(points)
    print(spts)
    draw(points)
    plt.show()

    # import ipdb; ipdb.set_trace()



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import

import numpy as np

from libs.configs._base_.models.retinanet_r50_fpn import *
from libs.configs._base_.datasets.dota_detection import *
from libs.configs._base_.schedules.schedule_1x import *
from dataloader.pretrained_weights.pretrain_zoo import PretrainModelZoo

# schedule
BATCH_SIZE = 1
GPU_GROUP = "0"
NUM_GPU = len(GPU_GROUP.strip().split(','))
SAVE_WEIGHTS_INTE = 27000
DECAY_STEP = np.array(DECAY_EPOCH, np.int32) * SAVE_WEIGHTS_INTE
MAX_ITERATION = SAVE_WEIGHTS_INTE * MAX_EPOCH
WARM_SETP = int(WARM_EPOCH * SAVE_WEIGHTS_INTE)

# model
# backbone
pretrain_zoo = PretrainModelZoo()
PRETRAINED_CKPT = pretrain_zoo.pretrain_weight_path(NET_NAME, ROOT_PATH)
TRAINED_CKPT = os.path.join(ROOT_PATH, 'output/trained_weights')

# bbox head
METHOD = 'R'
ANCHOR_RATIOS = [1, 1 / 3., 3.]

# loss
CLS_WEIGHT = 1.0
REG_WEIGHT = 1.0 / 5.0
REG_LOSS_MODE = None

VERSION = 'RetinaNet_DOTA_1x_20210610'

"""
RetinaNet-R + 90
FLOPs: 512292465;    Trainable params: 34524216
This is your result for task 1:

    mAP: 0.6725084704054188
    ap of each class:
    plane:0.8928344652701296,
    baseball-diamond:0.7304554873104306,
    bridge:0.34861749965443384,
    ground-track-field:0.6454757711323661,
    small-vehicle:0.7358328451641203,
    large-vehicle:0.7333202044070867,
    ship:0.829537451899966,
    tennis-court:0.8991725822927377,
    basketball-court:0.7414258616115545,
    storage-tank:0.7907933807503915,
    soccer-ball-field:0.5718970329039421,
    roundabout:0.5956995083477648,
    harbor:0.51027563083865,
    swimming-pool:0.6614448075148596,
    helicopter:0.4008445269828484

The submitted information is :

Description: RetinaNet_DOTA_1x_20210610_35.1w
Username: SJTU-Det
Institute: SJTU
Emailadress: yangxue-2019-sjtu@sjtu.edu.cn
TeamMembers: yangxue
"""




# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import

import numpy as np

from libs.configs._base_.models.retinanet_r50_fpn import *
from libs.configs._base_.datasets.dota_detection import *
from libs.configs._base_.schedules.schedule_1x import *
from dataloader.pretrained_weights.pretrain_zoo import PretrainModelZoo

# schedule
BATCH_SIZE = 1
GPU_GROUP = "0,1,2"
NUM_GPU = len(GPU_GROUP.strip().split(','))
SAVE_WEIGHTS_INTE = 27000 * 2
DECAY_STEP = np.array(DECAY_EPOCH, np.int32) * SAVE_WEIGHTS_INTE
MAX_ITERATION = SAVE_WEIGHTS_INTE * MAX_EPOCH
WARM_SETP = int(WARM_EPOCH * SAVE_WEIGHTS_INTE)

# dataset

# model
# backbone
pretrain_zoo = PretrainModelZoo()
PRETRAINED_CKPT = pretrain_zoo.pretrain_weight_path(NET_NAME, ROOT_PATH)
TRAINED_CKPT = os.path.join(ROOT_PATH, 'output/trained_weights')

# loss
CLS_WEIGHT = 1.0
REG_WEIGHT = 1.0 / 5.0
REG_LOSS_MODE = None

VERSION = 'RetinaNet_DOTA_2x_20190530'

"""
RetinaNet-H + 90
This is your result for task 1:

    mAP: 0.6572506703256068
    ap of each class:
    plane:0.8831119481871824,
    baseball-diamond:0.7554052281871614,
    bridge:0.4217303911789575,
    ground-track-field:0.6707230071220774,
    small-vehicle:0.6592650965532021,
    large-vehicle:0.5111005162900164,
    ship:0.7261407293679227,
    tennis-court:0.9071013790480128,
    basketball-court:0.7822207883168055,
    storage-tank:0.7883844023962553,
    soccer-ball-field:0.544082059014562,
    roundabout:0.6200017658693254,
    harbor:0.5324027345069116,
    swimming-pool:0.6718903394664805,
    helicopter:0.3851996693792289

The submitted information is :

Description: RetinaNet_DOTA_2x_20190530_108w
Username: DetectionTeamCSU
Institute: CSU
Emailadress: yangxue@csu.edu.cn
TeamMembers: YangXue

"""




# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import

import numpy as np

from libs.configs._base_.models.retinanet_r50_fpn import *
from libs.configs._base_.datasets.dota_detection import *
from libs.configs._base_.schedules.schedule_1x import *
from dataloader.pretrained_weights.pretrain_zoo import PretrainModelZoo

# schedule
BATCH_SIZE = 1
GPU_GROUP = "0,1,2"
NUM_GPU = len(GPU_GROUP.strip().split(','))
SAVE_WEIGHTS_INTE = 27000 * 2
DECAY_STEP = np.array(DECAY_EPOCH, np.int32) * SAVE_WEIGHTS_INTE
MAX_ITERATION = SAVE_WEIGHTS_INTE * MAX_EPOCH
WARM_SETP = int(WARM_EPOCH * SAVE_WEIGHTS_INTE)

# dataset

# model
# backbone
pretrain_zoo = PretrainModelZoo()
PRETRAINED_CKPT = pretrain_zoo.pretrain_weight_path(NET_NAME, ROOT_PATH)
TRAINED_CKPT = os.path.join(ROOT_PATH, 'output/trained_weights')

# bbox head
ANGLE_RANGE = 180

# loss
CLS_WEIGHT = 1.0
REG_WEIGHT = 1.0
ANGLE_WEIGHT = 0.5

# DCL
OMEGA = 180 / 256.  
ANGLE_MODE = 0  # {0: BCL, 1: GCL}

VERSION = 'RetinaNet_DOTA_DCL_B_2x_20200912'

"""
FLOPs: 872648361;    Trainable params: 33341751
v4 + period loss
This is your result for task 1:

    mAP: 0.673917075709117
    ap of each class:
    plane:0.8891746888406706,
    baseball-diamond:0.7211152849848342,
    bridge:0.41398424290638336,
    ground-track-field:0.6631986431727604,
    small-vehicle:0.6582194691343981,
    large-vehicle:0.5627183585938578,
    ship:0.7379739445312744,
    tennis-court:0.9078627422924039,
    basketball-court:0.7985916610693894,
    storage-tank:0.7902816236238415,
    soccer-ball-field:0.5411469473323773,
    roundabout:0.6024954269211565,
    harbor:0.5430105428794202,
    swimming-pool:0.6786164944083442,
    helicopter:0.6003660649456417

The submitted information is :

Description: RetinaNet_DOTA_DCL_B_2x_20200912_81w
Username: DetectionTeamCSU
Institute: CSU
Emailadress: yangxue@csu.edu.cn
TeamMembers: YangXue
"""




# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import

import numpy as np

from libs.configs._base_.models.retinanet_r50_fpn import *
from libs.configs._base_.datasets.dota_detection import *
from libs.configs._base_.schedules.schedule_1x import *
from dataloader.pretrained_weights.pretrain_zoo import PretrainModelZoo

# schedule
BATCH_SIZE = 1
GPU_GROUP = "0,1,2"
NUM_GPU = len(GPU_GROUP.strip().split(','))
SAVE_WEIGHTS_INTE = 27000 * 2
DECAY_STEP = np.array(DECAY_EPOCH, np.int32) * SAVE_WEIGHTS_INTE
MAX_ITERATION = SAVE_WEIGHTS_INTE * MAX_EPOCH
WARM_SETP = int(WARM_EPOCH * SAVE_WEIGHTS_INTE)

# dataset

# model
# backbone
pretrain_zoo = PretrainModelZoo()
PRETRAINED_CKPT = pretrain_zoo.pretrain_weight_path(NET_NAME, ROOT_PATH)
TRAINED_CKPT = os.path.join(ROOT_PATH, 'output/trained_weights')

# bbox head
ANGLE_RANGE = 180

# loss
CLS_WEIGHT = 1.0
REG_WEIGHT = 1.0
ANGLE_WEIGHT = 0.5

# DCL
OMEGA = 180 / 256.
ANGLE_MODE = 1  # {0: BCL, 1: GCL}

VERSION = 'RetinaNet_DOTA_DCL_G_2x_20201004'

"""
v3 + period loss

This is your result for task 1:

    mAP: 0.6701529258132946
    ap of each class:
    plane:0.8852753615354908,
    baseball-diamond:0.7358487312228768,
    bridge:0.397846525029383,
    ground-track-field:0.643821813715879,
    small-vehicle:0.67198526654463,
    large-vehicle:0.5602121060064053,
    ship:0.740990137738814,
    tennis-court:0.9078254427263777,
    basketball-court:0.7913191496695368,
    storage-tank:0.7794879186223984,
    soccer-ball-field:0.5560141958479111,
    roundabout:0.6190036754323164,
    harbor:0.5381769188233715,
    swimming-pool:0.6618352908439461,
    helicopter:0.5626513534400819

The submitted information is :

Description: RetinaNet_DOTA_DCL_G_2x_20201004_70.2w
Username: SJTU-Det
Institute: SJTU
Emailadress: yangxue-2019-sjtu@sjtu.edu.cn
TeamMembers: yangxue

"""



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys
import tensorflow as tf

sys.path.append('../..')
# from utils.gaussian_wasserstein_distance import get_element1, get_element4
from libs.utils.coordinate_convert import *
# from libs.utils.rbbox_overlaps import rbbx_overlaps
# from libs.utils.iou_cpu import get_iou_matrix


def iou_rotate_calculate2(boxes1, boxes2):
    ious = []
    if boxes1.shape[0] != 0:
        boxes1[:, 2] += 1.0
        boxes1[:, 3] += 1.0
        boxes2[:, 2] += 1.0
        boxes2[:, 3] += 1.0

        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]

        for i in range(boxes1.shape[0]):
            temp_ious = []
            r1 = ((boxes1[i][0], boxes1[i][1]), (boxes1[i][2], boxes1[i][3]), boxes1[i][4])
            r2 = ((boxes2[i][0], boxes2[i][1]), (boxes2[i][2], boxes2[i][3]), boxes2[i][4])

            int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
            if int_pts is not None:
                order_pts = cv2.convexHull(int_pts, returnPoints=True)

                int_area = cv2.contourArea(order_pts)

                inter = int_area * 1.0 / (area1[i] + area2[i] - int_area + 1e-4)

                # if boxes1[i][2] < 0.1 or boxes1[i][3] < 0.1 or boxes2[i][2] < 0.1 or boxes2[i][3] < 0.1:
                #     inter = 0

                inter = max(0.0, min(1.0, inter))

                temp_ious.append(inter)
            else:
                temp_ious.append(0.0)
            ious.append(temp_ious)

    return np.array(ious, dtype=np.float32)


def diou_rotate_calculate(boxes1, boxes2):

    if boxes1.shape[0] != 0:
        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]
        d = (boxes1[:, 0] - boxes2[:, 0]) ** 2 + (boxes1[:, 1] - boxes2[:, 1])

        boxes1_ = forward_convert(boxes1, with_label=False)
        boxes2_ = forward_convert(boxes2, with_label=False)

        xmin = np.minimum(np.min(boxes1_[:, 0::2]), np.min(boxes2_[:, 0::2]))
        xmax = np.maximum(np.max(boxes1_[:, 0::2]), np.max(boxes2_[:, 0::2]))
        ymin = np.minimum(np.min(boxes1_[:, 1::2]), np.min(boxes2_[:, 1::2]))
        ymax = np.maximum(np.max(boxes1_[:, 1::2]), np.max(boxes2_[:, 1::2]))

        c = (xmax - xmin) ** 2 + (ymax - ymin) ** 2
        ious = []
        for i in range(boxes1.shape[0]):
            r1 = ((boxes1[i][0], boxes1[i][1]), (boxes1[i][2], boxes1[i][3]), boxes1[i][4])
            r2 = ((boxes2[i][0], boxes2[i][1]), (boxes2[i][2], boxes2[i][3]), boxes2[i][4])

            int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
            if int_pts is not None:
                order_pts = cv2.convexHull(int_pts, returnPoints=True)

                int_area = cv2.contourArea(order_pts)

                iou = int_area * 1.0 / (area1[i] + area2[i] - int_area)
            else:
                iou = 0.0

            ious.append(iou)
        ious = np.array(ious)

        dious = ious - d / c
    else:
        dious = []

    return np.reshape(np.array(dious, dtype=np.float32), [-1, 1])


def gaussian_wasserstein_distance_(boxes1, boxes2):
    boxes1 = coordinate_present_convert(boxes1, -1)
    boxes1[:, 4] += 90
    boxes1[:, 4] *= (-np.pi / 180)

    boxes2 = coordinate_present_convert(boxes2, -1)
    boxes2[:, 4] += 90
    boxes2[:, 4] *= (-np.pi / 180)

    dis = (boxes1[:, 0] - boxes2[:, 0])**2 + (boxes1[:, 1] - boxes2[:, 1])**2 + \
          ((boxes1[:, 2] / 2 * np.cos(boxes1[:, 4])**2 + boxes1[:, 3] / 2 * np.sin(boxes1[:, 4])**2) - (boxes2[:, 2] / 2 * np.cos(boxes2[:, 4])**2 + boxes2[:, 3] / 2 * np.sin(boxes2[:, 4])**2))**2 + \
          2*((boxes1[:, 2] / 2 * np.cos(boxes1[:, 4])* np.sin(boxes1[:, 4] - boxes1[:, 3] / 2 * np.cos(boxes1[:, 4]) * np.sin(boxes1[:, 4]))) - (boxes2[:, 2] / 2 * np.cos(boxes2[:, 4]) * np.sin(boxes2[:, 4] - boxes2[:, 3] / 2 * np.cos(boxes2[:, 4]) * np.sin(boxes2[:, 4]))))**2 + \
          ((boxes1[:, 2] / 2 * np.sin(boxes1[:, 4]) ** 2 + boxes1[:, 3] / 2 * np.cos(boxes1[:, 4]) ** 2) - (
          boxes2[:, 2] / 2 * np.sin(boxes2[:, 4]) ** 2 + boxes2[:, 3] / 2 * np.cos(boxes2[:, 4]) ** 2)) ** 2
    return dis


def gaussian_wasserstein_distance(boxes1, boxes2):
    from utils.gaussian_wasserstein_distance import get_element1, get_element4

    boxes1 = coordinate_present_convert(boxes1, -1)
    boxes1[:, 4] += 90
    boxes1[:, 4] *= (-np.pi / 180)
    #
    boxes2 = coordinate_present_convert(boxes2, -1)
    boxes2[:, 4] += 90
    boxes2[:, 4] *= (-np.pi / 180)

    element1 = get_element1(boxes1[:, 2], boxes1[:, 3], boxes1[:, 4], boxes2[:, 2], boxes2[:, 3], boxes2[:, 4])
    element4 = get_element4(boxes1[:, 2], boxes1[:, 3], boxes1[:, 4], boxes2[:, 2], boxes2[:, 3], boxes2[:, 4])
    dis = (boxes1[:, 0] - boxes2[:, 0])**2 + (boxes1[:, 1] - boxes2[:, 1])**2 + (element1 + element4)
    return dis


def wasserstein_diss_sigma(sigma1, sigma2):
    wasserstein_diss_item2 = tf.linalg.matmul(sigma1, sigma1) + tf.linalg.matmul(sigma2, sigma2) - 2 * tf.linalg.sqrtm(
        tf.linalg.matmul(tf.linalg.matmul(sigma1, tf.linalg.matmul(sigma2, sigma2)), sigma1))
    wasserstein_diss_item2 = tf.linalg.trace(wasserstein_diss_item2)
    return wasserstein_diss_item2


def gwd(boxes1, boxes2):
    x1, y1, w1, h1, theta1 = tf.unstack(boxes1, axis=1)
    x2, y2, w2, h2, theta2 = tf.unstack(boxes2, axis=1)
    x1 = tf.reshape(x1, [-1, 1])
    y1 = tf.reshape(y1, [-1, 1])
    h1 = tf.reshape(h1, [-1, 1])
    w1 = tf.reshape(w1, [-1, 1])
    theta1 = tf.reshape(theta1, [-1, 1])
    x2 = tf.reshape(x2, [-1, 1])
    y2 = tf.reshape(y2, [-1, 1])
    h2 = tf.reshape(h2, [-1, 1])
    w2 = tf.reshape(w2, [-1, 1])
    theta2 = tf.reshape(theta2, [-1, 1])
    theta1 *= (np.pi / 180)
    theta2 *= (np.pi / 180)

    sigma1_1 = w1 / 2 * tf.cos(theta1) ** 2 + h1 / 2 * tf.sin(theta1) ** 2
    sigma1_2 = w1 / 2 * tf.sin(theta1) * tf.cos(theta1) - h1 / 2 * tf.sin(theta1) * tf.cos(theta1)
    sigma1_3 = w1 / 2 * tf.sin(theta1) * tf.cos(theta1) - h1 / 2 * tf.sin(theta1) * tf.cos(theta1)
    sigma1_4 = w1 / 2 * tf.sin(theta1) ** 2 + h1 / 2 * tf.cos(theta1) ** 2
    sigma1 = tf.reshape(tf.concat([sigma1_1, sigma1_2, sigma1_3, sigma1_4], axis=-1), [-1, 2, 2])

    sigma2_1 = w2 / 2 * tf.cos(theta2) ** 2 + h2 / 2 * tf.sin(theta2) ** 2
    sigma2_2 = w2 / 2 * tf.sin(theta2) * tf.cos(theta2) - h2 / 2 * tf.sin(theta2) * tf.cos(theta2)
    sigma2_3 = w2 / 2 * tf.sin(theta2) * tf.cos(theta2) - h2 / 2 * tf.sin(theta2) * tf.cos(theta2)
    sigma2_4 = w2 / 2 * tf.sin(theta2) ** 2 + h2 / 2 * tf.cos(theta2) ** 2
    sigma2 = tf.reshape(tf.concat([sigma2_1, sigma2_2, sigma2_3, sigma2_4], axis=-1), [-1, 2, 2])

    wasserstein_diss_item1 = (x1 - x2) ** 2 + (y1 - y2) ** 2
    wasserstein_diss_item2 = tf.reshape(wasserstein_diss_sigma(sigma1, sigma2), [-1, 1])
    wasserstein_diss = wasserstein_diss_item1 + wasserstein_diss_item2
    return sigma1, sigma2, wasserstein_diss


def sigma(a, w, h):
    R = np.array([[np.cos(a), -np.sin(a)], [np.sin(a), np.cos(a)]])
    sig = np.array([[w/2, 0], [0, h/2]])
    res = np.dot(R, sig)
    res = np.dot(res, R.T)
    return res


if __name__ == '__main__':
    from utils.gaussian_wasserstein_distance import get_element1, get_element4

    boxes1 = np.array([[50, 50, 10, 70, -30],
                       [50, 50, 100, 700, -30]], np.float32)

    boxes2 = np.array([[10, 40, 10, 70, -30],
                       [10, 40, 100, 700, -40]], np.float32)

    # boxes1 = np.array([    # prediction box
    #     [50, 50, 10, 70, -35],     # 90 <--> 180
    #     [50, 50, 70, 10, -90.5],   # 90   PoA + EoE
    #     [50, 50, 70, 10, -90.5],   # 180  PoA
    #     [50, 50, 40, 40, -35],     # 180  w=h
    # ], np.float32)
    #
    # boxes2 = np.array([    # ground truth
    #     [50, 50, 70, 10, 55],
    #     [50, 50, 10, 70, -0.5],
    #     [50, 50, 70, 10, 89.5],
    #     [50, 50, 40, 40, 55],
    # ], np.float32)

    print('iou', iou_rotate_calculate2(boxes1, boxes2).reshape(-1,))  # [0.9999996 0.9999998 0.9999998 1.       ]
    # print(diou_rotate_calculate(boxes1, boxes2).reshape(-1,))  # [0.9999997  0.99999994 0.99999994 1.        ]
    # print(gaussian_wasserstein_distance(boxes1, boxes2))     # [6.1035156e-05 3.1062821e-04 3.1062821e-04 0.0000000e+00]


    # tmp = np.maximum(np.log(gaussian_wasserstein_distance(boxes1, boxes2)+1e-3), 0)
    # print(np.log(gaussian_wasserstein_distance(boxes1, boxes2)))
    # print(tmp/(1+tmp))

    # print(np.argsort(iou_rotate_calculate2(boxes1, boxes2).reshape(-1, )*-1))
    # print(np.argsort(diou_rotate_calculate(boxes1, boxes2).reshape(-1, )*-1))
    # print(np.argsort(np.array(gaussian_wasserstein_distance(boxes1, boxes2))))
    #
    # # print(sigma(-np.pi*35/180, 10, 70))
    # # print(sigma(np.pi*(90-35)/180, 70, 10))
    #
    sigma1_tf1, sigma2_tf1, gwd_tf1 = gwd(coordinate_present_convert(boxes1, -1, False), coordinate_present_convert(boxes2, -1, False))
    sigma1_tf2, sigma2_tf2, gwd_tf2 = gwd(boxes1, boxes2)
    # sigma1_tf2 = tf.linalg.eigvalsh(sigma1_tf2)

    with tf.Session() as sess:
        sigma1_tf_1, sigma2_tf_1, gwd_tf_1 = sess.run([sigma1_tf1, sigma2_tf1, gwd_tf1])
        sigma1_tf_2, sigma2_tf_2, gwd_tf_2 = sess.run([sigma1_tf2, sigma2_tf2, gwd_tf2])
        # print(sigma1_tf_1)
        # print(sigma2_tf_1)
        # print('**'*10)
        print(sigma1_tf_2)
        # print(sigma2_tf_2)

        print('**' * 10)
        # print(np.reshape(gwd_tf_1, [-1, ]))
        # print(np.argsort(np.reshape(gwd_tf_1, [-1, ])))

        print('gwd', np.reshape(gwd_tf_2, [-1, ]))
        print(1-1/(np.sqrt(gwd_tf_2)+1))
        # print(np.argsort(np.reshape(gwd_tf_2, [-1, ])))

        # gwd_tf_2 = np.maximum(np.log(gwd_tf_2 + 1e-3), 0.0)
        # gwd_tf_2_ = np.maximum(np.log(gwd_tf_2 + 1e-3), 0.0)
        # print(gwd_tf_2_/(5+gwd_tf_2_))
        #
        # gwd_tf_2 = np.maximum(np.log(gwd_tf_2 + 1e-3), 0.0)
        # print(gwd_tf_2)
        # print(1-1 / (5 + gwd_tf_2))
        # print(gwd_tf_2_ / (2 + gwd_tf_2_))
        # print(gwd_tf_2_ / (3 + gwd_tf_2_))
        # print(gwd_tf_2_ / (5 + gwd_tf_2_))










# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import

import numpy as np

from libs.configs._base_.models.retinanet_r50_fpn import *
from libs.configs._base_.datasets.dota_detection import *
from libs.configs._base_.schedules.schedule_1x import *
from dataloader.pretrained_weights.pretrain_zoo import PretrainModelZoo

# schedule
BATCH_SIZE = 1
GPU_GROUP = "0"
NUM_GPU = len(GPU_GROUP.strip().split(','))
LR = 1e-3
SAVE_WEIGHTS_INTE = 27000
DECAY_STEP = np.array(DECAY_EPOCH, np.int32) * SAVE_WEIGHTS_INTE
MAX_ITERATION = SAVE_WEIGHTS_INTE * MAX_EPOCH
WARM_EPOCH = 1. / 8.
WARM_SETP = int(WARM_EPOCH * SAVE_WEIGHTS_INTE)

# dataset

# model
pretrain_zoo = PretrainModelZoo()
PRETRAINED_CKPT = pretrain_zoo.pretrain_weight_path(NET_NAME, ROOT_PATH)
TRAINED_CKPT = os.path.join(ROOT_PATH, 'output/trained_weights')

# loss
CLS_WEIGHT = 1.0
REG_WEIGHT = 2.0
REG_LOSS_MODE = 3  # KLD loss

KL_TAU = 2.0
KL_FUNC = 0   # 0: sqrt  1: log

VERSION = 'RetinaNet_DOTA_KL_1x_20210201'

"""
RetinaNet-H + kl (fix bug) + sqrt + tau=2
FLOPs: 484911761;    Trainable params: 33002916
This is your result for task 1:

mAP: 0.7128325571761713
ap of each class:
plane:0.884289204525325,
baseball-diamond:0.7653565915743398,
bridge:0.440047662898936,
ground-track-field:0.698238996872059,
small-vehicle:0.7444580421686285,
large-vehicle:0.7248184249702364,
ship:0.843025375274411,
tennis-court:0.8939539261877734,
basketball-court:0.806571747424402,
storage-tank:0.800303800899483,
soccer-ball-field:0.5787146175790521,
roundabout:0.6505316568373755,
harbor:0.6554363687620437,
swimming-pool:0.6686981948609415,
helicopter:0.5380437468075643

The submitted information is :

Description: RetinaNet_DOTA_KL_1x_20210201_45.9w
Username: SJTU-Det
Institute: SJTU
Emailadress: yangxue-2019-sjtu@sjtu.edu.cn
TeamMembers: yangxue
"""



import numpy as np
import cv2


def points_sampling(mask, num=4):
    mask = np.array(mask)

    # mask = mask.reshape((-1, 1, 2))
    # mask = cv2.convexHull(mask, returnPoints=True)
    # mask = mask.reshape((-1, 2))

    pnum, cnum = mask.shape
    assert cnum == 2
    idxnext_p = (np.arange(pnum, dtype=np.int32) + 1) % pnum
    pgtnext_px2 = mask[idxnext_p]
    edgelen_p = np.sqrt(np.sum((pgtnext_px2 - mask) ** 2, axis=1))
    edgeidxsort_p = np.argsort(edgelen_p)

    # two cases
    # we need to remove gt points
    # we simply remove shortest paths
    if pnum > num:
        edgeidxkeep_k = edgeidxsort_p[pnum - num:]
        edgeidxsort_k = np.sort(edgeidxkeep_k)
        pgtnp_kx2 = mask[edgeidxsort_k]
        assert pgtnp_kx2.shape[0] == num
        return pgtnp_kx2
    # we need to add gt points
    # we simply add it uniformly
    else:
        edgenum = np.round(edgelen_p * num / np.sum(edgelen_p)).astype(np.int32)
        for i in range(pnum):
            if edgenum[i] == 0:
                edgenum[i] = 1

        # after round, it may has 1 or 2 mismatch
        edgenumsum = np.sum(edgenum)
        if edgenumsum != num:

            if edgenumsum > num:

                id = -1
                passnum = edgenumsum - num
                while passnum > 0:
                    edgeid = edgeidxsort_p[id]
                    if edgenum[edgeid] > passnum:
                        edgenum[edgeid] -= passnum
                        passnum -= passnum
                    else:
                        passnum -= edgenum[edgeid] - 1
                        edgenum[edgeid] -= edgenum[edgeid] - 1
                        id -= 1
            else:
                id = -1
                edgeid = edgeidxsort_p[id]
                edgenum[edgeid] += num - edgenumsum

        assert np.sum(edgenum) == num

        psample = []
        for i in range(pnum):
            pb_1x2 = mask[i:i + 1]
            pe_1x2 = pgtnext_px2[i:i + 1]

            pnewnum = edgenum[i]
            wnp_kx1 = np.arange(edgenum[i], dtype=np.float32).reshape(-1, 1) / edgenum[i]

            pmids = pb_1x2 * (1 - wnp_kx1) + pe_1x2 * wnp_kx1
            psample.append(pmids)

        psamplenp = np.concatenate(psample, axis=0)
        return psamplenp


def mask_sampling(masks, num=8):
    return np.array([points_sampling(m, num) for m in masks], np.float32)

if __name__ == '__main__':
    mask1 = np.array([[[10, 0], [5, 10], [0, 10], [10, 5], [0, 0], [5, 0], [10, 10], [0, 5]],
                      [[0, 0], [5, 0], [10, 0], [10, 5], [10, 10], [5, 10], [0, 10], [0, 5]]])
    mask2 = np.array([[[0, 0], [10, 0], [10, 10], [0, 10]]])

    points = mask_sampling(mask2, num=16)
    print(points)



# -*- coding: utf-8 -*-
from __future__ import division, print_function, absolute_import
import math
import sys
import os
import cv2
import numpy as np
# import tfplot as tfp

from libs.utils.coordinate_convert import forward_convert


def view_bar(message, num, total):
    rate = num / total
    rate_num = int(rate * 40)
    rate_nums = math.ceil(rate * 100)
    r = '\r%s:[%s%s]%d%%\t%d/%d' % (message, ">" * rate_num, " " * (40 - rate_num), rate_nums, num, total,)
    sys.stdout.write(r)
    sys.stdout.flush()


def makedirs(path):
    if not os.path.exists(path):
        os.makedirs(path)


def get_dota_short_names(label):
    DOTA_SHORT_NAMES = {
        'roundabout': 'RA',
        'tennis-court': 'TC',
        'swimming-pool': 'SP',
        'storage-tank': 'ST',
        'soccer-ball-field': 'SBF',
        'small-vehicle': 'SV',
        'ship': 'SH',
        'plane': 'PL',
        'large-vehicle': 'LV',
        'helicopter': 'HC',
        'harbor': 'HA',
        'ground-track-field': 'GTF',
        'bridge': 'BR',
        'basketball-court': 'BC',
        'baseball-diamond': 'BD',
        'container-crane': 'CC',
        'airport': 'AP',
        'helipad': 'HP'
    }

    return DOTA_SHORT_NAMES[label]


def read_dota_gt_and_vis(img, gt_txt):
    txt_data = open(gt_txt, 'r').readlines()
    for i in txt_data:
        if len(i.split(' ')) < 9:
            continue

        gt_box = [int(xy) for xy in i.split(' ')[:8]]
        # gt_label = i.split(' ')[8]
        cv2.line(img, (gt_box[0], gt_box[1]), (gt_box[2], gt_box[3]), color=(0, 0, 255), thickness=3)
        cv2.line(img, (gt_box[2], gt_box[3]), (gt_box[4], gt_box[5]), color=(0, 0, 255), thickness=3)
        cv2.line(img, (gt_box[4], gt_box[5]), (gt_box[6], gt_box[7]), color=(0, 0, 255), thickness=3)
        cv2.line(img, (gt_box[6], gt_box[7]), (gt_box[0], gt_box[1]), color=(0, 0, 255), thickness=3)
    return img


def get_mask(img, boxes):
    boxes = forward_convert(boxes)
    h, w, _ = img.shape
    mask = np.zeros([h, w])
    for b in boxes:
        b = np.reshape(b[0:-1], [4, 2])
        rect = np.array(b, np.int32)
        cv2.fillConvexPoly(mask, rect, 1)
    # mask = cv2.resize(mask, dsize=(h // 16, w // 16))
    mask = np.expand_dims(mask, axis=-1)
    return np.array(mask, np.float32)


# def add_heatmap(feature_maps, name):
#     '''
#     :param feature_maps:[B, H, W, C]
#     :return:
#     '''
#
#     def figure_attention(activation):
#         fig, ax = tfp.subplots()
#         im = ax.imshow(activation, cmap='jet')
#         fig.colorbar(im)
#         return fig
#
#     heatmap = tf.reduce_sum(feature_maps, axis=-1)
#     heatmap = tf.squeeze(heatmap, axis=0)
#     tfp.summary.plot(name, figure_attention, [heatmap])



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

import numpy as np


def hbox_grid_sample(boxes, point_num_per_line=3):

    bin_w, bin_h = (boxes[:, 2] - boxes[:, 0]) / (point_num_per_line-1), (boxes[:, 3] - boxes[:, 1]) / (point_num_per_line-1)

    shift_x = np.expand_dims(np.arange(0, point_num_per_line), axis=0)
    shift_y = np.expand_dims(np.arange(0, point_num_per_line), axis=0)
    shift_x, shift_y = np.meshgrid(shift_x, shift_y)
    shift_x = np.reshape(shift_x, [-1, 1])
    shift_y = np.reshape(shift_y, [-1, 1])
    shifts = np.concatenate([shift_x, shift_y], axis=-1)
    shifts = np.reshape(shifts, [-1, point_num_per_line**2*2])
    shifts = np.array(np.tile(shifts, [boxes.shape[0], 1]).reshape([-1, point_num_per_line**2, 2]), np.float32)
    bin_w = np.reshape(bin_w, [-1, 1])
    bin_h = np.reshape(bin_h, [-1, 1])
    shifts[:, :, 0] *= bin_w
    shifts[:, :, 1] *= bin_h
    shifts += boxes[:, np.newaxis, 0:2]
    return shifts.reshape([-1, point_num_per_line**2*2])


def rbox_border_sample(boxes, point_num_per_line=3):
    shift_x = np.arange(0, point_num_per_line-1)
    shift_y = np.arange(0, point_num_per_line-1)
    shift_x = np.reshape(shift_x, [-1, 1])
    shift_y = np.reshape(shift_y, [-1, 1])

    sample_points_list = []
    for i in range(4):
        width = boxes[:, (i*2+2)%8] - boxes[:, (i*2)%8]
        height = boxes[:, (i*2+3)%8] - boxes[:, (i*2+1)%8]
        bin_w, bin_h = width / (point_num_per_line-1), height / (point_num_per_line-1)

        shifts = np.concatenate([shift_x, shift_y], axis=-1)
        shifts = np.array(np.tile(shifts, [boxes.shape[0], 1]).reshape([-1, point_num_per_line-1, 2]), np.float32)
        bin_w = np.reshape(bin_w, [-1, 1])
        bin_h = np.reshape(bin_h, [-1, 1])
        shifts[:, :, 0] *= bin_w
        shifts[:, :, 1] *= bin_h
        shifts += boxes[:, np.newaxis, i*2:(i*2)+2]
        sample_points_list.append(shifts)
    sample_points = np.concatenate(sample_points_list, axis=1)
    return sample_points.reshape([-1, (point_num_per_line - 1) * 4 * 2])


if __name__ == '__main__':
    # print(hbox_grid_sample(np.array([[3, 3, 12, 12], [0, 0, 4, 4]]), 3))
    print(rbox_border_sample(np.array([[3, 3, 12, 3, 12, 12, 3, 12], [0, 0, 4, 0, 4, 4, 0, 4]]), 3))



import numpy as np

from distutils.core import setup, Extension
from Cython.Build import cythonize

try:
    numpy_include = np.get_include()
except AttributeError:
    numpy_include = np.get_numpy_include()

setup(ext_modules=cythonize(Extension(
    "cpu_nms",
    ["cpu_nms.pyx"],
    language='c++',
    extra_compile_args=['-std=c++11'],
    include_dirs=[numpy_include]))
)



import cv2
import os
import numpy as np

img_path = '/data/dataset/DOTA/test/images/'
npy_path = '/data/dataset/DOTA/test/npy/'
if not os.path.exists(npy_path):
    os.makedirs(npy_path)
images = os.listdir(img_path)

for i in images:
    img = cv2.imread(os.path.join(img_path, i))
    print(img.shape)
    np.save(os.path.join(npy_path, i.replace('.png', '.npy')), img)
    img_npy = np.load(os.path.join(npy_path, i.replace('.png', '.npy')))
    print(img_npy.shape)



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import sys

sys.path.append('../..')
from libs.utils.coordinate_convert import *
from libs.utils.rbbox_overlaps import rbbx_overlaps
from libs.utils.iou_cpu import get_iou_matrix


def iou_rotate_calculate(boxes1, boxes2, use_gpu=True, gpu_id=0):
    '''

    :param boxes_list1:[N, 8] tensor
    :param boxes_list2: [M, 8] tensor
    :return:
    '''

    boxes1 = tf.cast(boxes1, tf.float32)
    boxes2 = tf.cast(boxes2, tf.float32)
    if use_gpu:

        iou_matrix = tf.py_func(rbbx_overlaps,
                                inp=[boxes1, boxes2, gpu_id],
                                Tout=tf.float32)
    else:
        iou_matrix = tf.py_func(get_iou_matrix, inp=[boxes1, boxes2],
                                Tout=tf.float32)

    iou_matrix = tf.reshape(iou_matrix, [tf.shape(boxes1)[0], tf.shape(boxes2)[0]])

    return iou_matrix


def iou_rotate_calculate1(boxes1, boxes2, use_gpu=True, gpu_id=0):

    # start = time.time()
    if use_gpu:
        ious = rbbx_overlaps(boxes1, boxes2, gpu_id)
    else:
        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]
        ious = []
        for i, box1 in enumerate(boxes1):
            temp_ious = []
            r1 = ((box1[0], box1[1]), (box1[2], box1[3]), box1[4])
            for j, box2 in enumerate(boxes2):
                r2 = ((box2[0], box2[1]), (box2[2], box2[3]), box2[4])

                int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
                if int_pts is not None:
                    order_pts = cv2.convexHull(int_pts, returnPoints=True)

                    int_area = cv2.contourArea(order_pts)

                    inter = int_area * 1.0 / (area1[i] + area2[j] - int_area + 1e-5)
                    temp_ious.append(inter)
                else:
                    temp_ious.append(0.0)
            ious.append(temp_ious)

    # print('{}s'.format(time.time() - start))

    return np.array(ious, dtype=np.float32)


def iou_rotate_calculate2(boxes1, boxes2):
    ious = []
    if boxes1.shape[0] != 0:
        boxes1[:, 2] += 1.0
        boxes1[:, 3] += 1.0
        boxes2[:, 2] += 1.0
        boxes2[:, 3] += 1.0

        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]

        for i in range(boxes1.shape[0]):
            temp_ious = []
            r1 = ((boxes1[i][0], boxes1[i][1]), (boxes1[i][2], boxes1[i][3]), boxes1[i][4])
            r2 = ((boxes2[i][0], boxes2[i][1]), (boxes2[i][2], boxes2[i][3]), boxes2[i][4])

            int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
            if int_pts is not None:
                order_pts = cv2.convexHull(int_pts, returnPoints=True)

                int_area = cv2.contourArea(order_pts)

                inter = int_area * 1.0 / (area1[i] + area2[i] - int_area + 1e-4)

                # if boxes1[i][2] < 0.1 or boxes1[i][3] < 0.1 or boxes2[i][2] < 0.1 or boxes2[i][3] < 0.1:
                #     inter = 0

                inter = max(0.0, min(1.0, inter))

                temp_ious.append(inter)
            else:
                temp_ious.append(0.0)
            ious.append(temp_ious)

    return np.array(ious, dtype=np.float32)


def diou_rotate_calculate(boxes1, boxes2):

    if boxes1.shape[0] != 0:
        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]
        d = (boxes1[:, 0] - boxes2[:, 0]) ** 2 + (boxes1[:, 1] - boxes2[:, 1])

        boxes1_ = forward_convert(boxes1, with_label=False)
        boxes2_ = forward_convert(boxes2, with_label=False)

        xmin = np.minimum(np.min(boxes1_[:, 0::2]), np.min(boxes2_[:, 0::2]))
        xmax = np.maximum(np.max(boxes1_[:, 0::2]), np.max(boxes2_[:, 0::2]))
        ymin = np.minimum(np.min(boxes1_[:, 1::2]), np.min(boxes2_[:, 1::2]))
        ymax = np.maximum(np.max(boxes1_[:, 1::2]), np.max(boxes2_[:, 1::2]))

        c = (xmax - xmin) ** 2 + (ymax - ymin) ** 2
        ious = []
        for i in range(boxes1.shape[0]):
            r1 = ((boxes1[i][0], boxes1[i][1]), (boxes1[i][2], boxes1[i][3]), boxes1[i][4])
            r2 = ((boxes2[i][0], boxes2[i][1]), (boxes2[i][2], boxes2[i][3]), boxes2[i][4])

            int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
            if int_pts is not None:
                order_pts = cv2.convexHull(int_pts, returnPoints=True)

                int_area = cv2.contourArea(order_pts)

                iou = int_area * 1.0 / (area1[i] + area2[i] - int_area)
            else:
                iou = 0.0

            ious.append(iou)
        ious = np.array(ious)

        dious = ious - d / c
    else:
        dious = []

    return np.reshape(np.array(dious, dtype=np.float32), [-1, 1])


def adiou_rotate_calculate(boxes1, boxes2):

    if boxes1.shape[0] != 0:
        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]
        d = (boxes1[:, 0] - boxes2[:, 0]) ** 2 + (boxes1[:, 1] - boxes2[:, 1])

        boxes1_ = forward_convert(boxes1, with_label=False)
        boxes2_ = forward_convert(boxes2, with_label=False)

        xmin = np.minimum(np.min(boxes1_[:, 0::2]), np.min(boxes2_[:, 0::2]))
        xmax = np.maximum(np.max(boxes1_[:, 0::2]), np.max(boxes2_[:, 0::2]))
        ymin = np.minimum(np.min(boxes1_[:, 1::2]), np.min(boxes2_[:, 1::2]))
        ymax = np.maximum(np.max(boxes1_[:, 1::2]), np.max(boxes2_[:, 1::2]))

        c = (xmax - xmin) ** 2 + (ymax - ymin) ** 2

        # v = (4 / (math.pi ** 2)) * (np.arctan(boxes1[:, 2]/boxes1[:, 3]) - np.arctan(boxes2[:, 2]/boxes2[:, 3])) ** 2

        ious = []
        for i in range(boxes1.shape[0]):
            r1 = ((boxes1[i][0], boxes1[i][1]), (boxes1[i][2], boxes1[i][3]), boxes1[i][4])
            r2 = ((boxes2[i][0], boxes2[i][1]), (boxes2[i][2], boxes2[i][3]), boxes2[i][4])

            int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
            if int_pts is not None:
                order_pts = cv2.convexHull(int_pts, returnPoints=True)

                int_area = cv2.contourArea(order_pts)

                iou = int_area * 1.0 / (area1[i] + area2[i] - int_area)
            else:
                iou = 0.0

            ious.append(iou)
        ious = np.array(ious)

        # S = 1 - ious
        # alpha = v / (S + v)
        # w_temp = 2 * boxes1[:, 2]
        # ar = (8 / (math.pi ** 2)) * (np.arctan(boxes1[:, 2]/boxes1[:, 3]) - np.arctan(boxes2[:, 2]/boxes2[:, 3])) \
        #      * ((boxes1[:, 2] - w_temp) * boxes1[:, 3])
        # cious = ious - d / c - alpha * ar
        cious = (ious - d / c) * np.abs(np.cos(boxes1[:, 4] - boxes2[:, 4]))
    else:
        cious = []

    return np.reshape(np.array(cious, dtype=np.float32), [-1, 1])

if __name__ == '__main__':

    boxes1 = np.array([
        [150, 150, 1e-15, 5, -45],
    ], np.float32)

    boxes2 = np.array([
        [50, 50, 5, 1e-15, -45]], np.float32)

    from libs.box_utils.overlaps_cython.overlaps_cython import rbox_overlaps

    overlaps = rbox_overlaps(np.array(boxes1, dtype=np.float32),
                             np.array(boxes2, dtype=np.float32))

    print(overlaps)

    print(rbbx_overlaps(boxes1, boxes2, 3))







import numpy as np
import math
import tensorflow as tf
import sys
sys.path.append('../')
from libs.utils.coordinate_convert import backward_convert


def dist(p1, p2):
    return np.linalg.norm(p1 - p2, axis=1)


def quad2rbox(boxes):
    # input: ordered points (bs, 8) 
    nB = len(boxes)
    points = boxes.reshape(-1, 4, 2)
    cxs = points[:, :, 0].sum(1)[:, np.newaxis] / 4
    cys = points[:, :, 1].sum(1)[:, np.newaxis] / 4
    _ws = dist(points[:, 0], points[:, 1])[:, np.newaxis]
    _hs = dist(points[:, 1], points[:, 2])[:, np.newaxis]
    # adjust theta
    _thetas = np.arctan2(-(points[:, 1, 0] - points[:, 0, 0]), points[:, 1, 1] - points[:, 0, 1])[:, np.newaxis]
    odd = (np.mod((_thetas // (-np.pi * 0.5)), 2) == 0)
    ws = np.where(odd, _hs, _ws)
    hs = np.where(odd, _ws, _hs)
    thetas = np.mod(_thetas, -np.pi * 0.5)
    rboxes = np.concatenate([cxs, cys, ws, hs, thetas / np.pi * 180.], 1)

    return rboxes


def dist_tf(p1, p2):
    return tf.linalg.norm(p1 - p2, axis=-1)


def quad2rbox_tf(boxes):
    points = tf.reshape(boxes, [-1, 4, 2])
    cxs = tf.expand_dims(tf.reduce_sum(points[:, :, 0], axis=1), axis=1) / 4.
    cys = tf.expand_dims(tf.reduce_sum(points[:, :, 1], axis=1), axis=1) / 4.
    _ws = tf.expand_dims(dist_tf(points[:, 0], points[:, 1]), axis=1)
    _hs = tf.expand_dims(dist_tf(points[:, 1], points[:, 2]), axis=1)
    _thetas = tf.expand_dims(tf.atan2(-(points[:, 1, 0]-points[:, 0, 0]), points[:, 1, 1]-points[:, 0, 1]), axis=1)
    odd = tf.equal(tf.mod((_thetas // (-np.pi * 0.5)), 2), 0)
    ws = tf.where(odd, _hs, _ws)
    hs = tf.where(odd, _ws, _hs)
    thetas = tf.mod(_thetas, -np.pi * 0.5)
    rboxes = tf.concat([cxs, cys, ws, hs, thetas / np.pi * 180.], axis=1)
    return rboxes


if __name__ == "__main__":
    quad = np.array([[278, 418, 308, 331, 761, 581, 691, 668],
                     [758, 418, 348, 331, 241, 581, 591, 668],
                     [624, 112, 490, 93, 496, 50, 630, 68],
                     [10, 0, 30, 20, 20, 30, 0, 10]], np.float32)
    box = quad2rbox(quad)
    print(box)
    box_tf = quad2rbox_tf(quad)
    with tf.Session() as sess:
        box_tf_ = sess.run(box_tf)
        print(box_tf_)

        print(backward_convert(quad, False))





import cv2
import numpy as np


def hiou(boxes_1, boxes_2):

    """
    Calculate the Axis-aligned IoU

    :param boxes_1: horizontal bounding box1
    :param boxes_2: horizontal bounding box2
    :return: axis-aligned IoU
    """

    xmin_1, ymin_1, xmax_1, ymax_1 = np.split(boxes_1, 4, axis=1)

    xmin_2, ymin_2, xmax_2, ymax_2 = boxes_2[:, 0], boxes_2[:, 1], boxes_2[:, 2], boxes_2[:, 3]

    max_xmin = np.maximum(xmin_1, xmin_2)
    min_xmax = np.minimum(xmax_1, xmax_2)

    max_ymin = np.maximum(ymin_1, ymin_2)
    min_ymax = np.minimum(ymax_1, ymax_2)

    overlap_h = np.maximum(0., min_ymax - max_ymin)  # avoid h < 0
    overlap_w = np.maximum(0., min_xmax - max_xmin)

    overlaps = overlap_h * overlap_w

    area_1 = (xmax_1 - xmin_1) * (ymax_1 - ymin_1)  # [N, 1]
    area_2 = (xmax_2 - xmin_2) * (ymax_2 - ymin_2)  # [M, ]

    iou = overlaps / (area_1 + area_2 - overlaps)

    return iou


def riou(boxes1, boxes2):
    """
    Calculate the Skew IoU

    :param boxes_1: rotated bounding box1
    :param boxes_2: rotated bounding box2
    :return: Skew IoU
    """

    ious = []
    if boxes1.shape[0] != 0:
        boxes1[:, 2] += 1.0
        boxes1[:, 3] += 1.0
        boxes2[:, 2] += 1.0
        boxes2[:, 3] += 1.0

        area1 = boxes1[:, 2] * boxes1[:, 3]
        area2 = boxes2[:, 2] * boxes2[:, 3]

        for i in range(boxes1.shape[0]):
            temp_ious = []
            r1 = ((boxes1[i][0], boxes1[i][1]), (boxes1[i][2], boxes1[i][3]), boxes1[i][4])
            r2 = ((boxes2[i][0], boxes2[i][1]), (boxes2[i][2], boxes2[i][3]), boxes2[i][4])

            int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]
            if int_pts is not None:
                order_pts = cv2.convexHull(int_pts, returnPoints=True)

                int_area = cv2.contourArea(order_pts)

                inter = int_area * 1.0 / (area1[i] + area2[i] - int_area + 1e-4)

                # if boxes1[i][2] < 0.1 or boxes1[i][3] < 0.1 or boxes2[i][2] < 0.1 or boxes2[i][3] < 0.1:
                #     inter = 0

                inter = max(0.0, min(1.0, inter))

                temp_ious.append(inter)
            else:
                temp_ious.append(0.0)
            ious.append(temp_ious)

    return np.array(ious, dtype=np.float32)


if __name__ == '__main__':
    boxes1 = np.array([
        [50, 50, 70, 10, -45]], np.float32)

    boxes2 = np.array([
        [50, 25, 70, 10, -45]], np.float32)

    iou_r = riou(boxes1, boxes2)
    print(iou_r)

    from libs.utils.coordinate_convert import forward_convert
    boxes1 = forward_convert(boxes1, False)
    boxes2 = forward_convert(boxes2, False)

    x1_min = np.min(boxes1[:, ::2], axis=-1)
    y1_min = np.min(boxes1[:, 1::2], axis=-1)
    x1_max = np.max(boxes1[:, ::2], axis=-1)
    y1_max = np.max(boxes1[:, 1::2], axis=-1)
    boxes1 = np.transpose(np.stack([x1_min, y1_min, x1_max, y1_max]))

    x2_min = np.min(boxes2[:, ::2], axis=-1)
    y2_min = np.min(boxes2[:, 1::2], axis=-1)
    x2_max = np.max(boxes2[:, ::2], axis=-1)
    y2_max = np.max(boxes2[:, 1::2], axis=-1)
    boxes2 = np.transpose(np.stack([x2_min, y2_min, x2_max, y2_max]))

    iou_h = hiou(boxes1, boxes2)
    print(iou_h)




# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


import tensorflow as tf
import numpy as np


def iou_calculate(boxes_1, boxes_2):

    with tf.name_scope('iou_caculate'):

        xmin_1, ymin_1, xmax_1, ymax_1 = tf.unstack(boxes_1, axis=1)  # ymin_1 shape is [N, 1]..

        xmin_2, ymin_2, xmax_2, ymax_2 = tf.unstack(boxes_2, axis=1)  # ymin_2 shape is [M, ]..

        max_xmin = tf.maximum(xmin_1, xmin_2)
        min_xmax = tf.minimum(xmax_1, xmax_2)

        max_ymin = tf.maximum(ymin_1, ymin_2)
        min_ymax = tf.minimum(ymax_1, ymax_2)

        overlap_h = tf.maximum(0., min_ymax - max_ymin)  # avoid h < 0
        overlap_w = tf.maximum(0., min_xmax - max_xmin)

        overlaps = overlap_h * overlap_w

        area_1 = (xmax_1 - xmin_1) * (ymax_1 - ymin_1)  # [N, 1]
        area_2 = (xmax_2 - xmin_2) * (ymax_2 - ymin_2)  # [M, ]

        iou = overlaps / (area_1 + area_2 - overlaps)

        return iou


def iou_calculate_np(boxes_1, boxes_2):
    xmin_1, ymin_1, xmax_1, ymax_1 = np.split(boxes_1, 4, axis=1)
    # xmin_1, ymin_1, xmax_1, ymax_1 = boxes_1[:, 0], boxes_1[:, 1], boxes_1[:, 2], boxes_1[:, 3]

    xmin_2, ymin_2, xmax_2, ymax_2 = boxes_2[:, 0], boxes_2[:, 1], boxes_2[:, 2], boxes_2[:, 3]

    max_xmin = np.maximum(xmin_1, xmin_2)
    min_xmax = np.minimum(xmax_1, xmax_2)

    max_ymin = np.maximum(ymin_1, ymin_2)
    min_ymax = np.minimum(ymax_1, ymax_2)

    overlap_h = np.maximum(0., min_ymax - max_ymin)  # avoid h < 0
    overlap_w = np.maximum(0., min_xmax - max_xmin)

    overlaps = overlap_h * overlap_w

    area_1 = (xmax_1 - xmin_1) * (ymax_1 - ymin_1)  # [N, 1]
    area_2 = (xmax_2 - xmin_2) * (ymax_2 - ymin_2)  # [M, ]

    iou = overlaps / (area_1 + area_2 - overlaps)

    return iou


def iou_calculate1(boxes_1, boxes_2):

    xmin_1, ymin_1, xmax_1, ymax_1 = boxes_1[:, 0], boxes_1[:, 1], boxes_1[:, 2], boxes_1[:, 3]

    xmin_2, ymin_2, xmax_2, ymax_2 = boxes_2[:, 0], boxes_2[:, 1], boxes_2[:, 2], boxes_2[:, 3]

    max_xmin = np.maximum(xmin_1, xmin_2)
    min_xmax = np.minimum(xmax_1, xmax_2)

    max_ymin = np.maximum(ymin_1, ymin_2)
    min_ymax = np.minimum(ymax_1, ymax_2)

    overlap_h = np.maximum(0., min_ymax - max_ymin)  # avoid h < 0
    overlap_w = np.maximum(0., min_xmax - max_xmin)

    overlaps = overlap_h * overlap_w

    area_1 = (xmax_1 - xmin_1) * (ymax_1 - ymin_1)  # [N, 1]
    area_2 = (xmax_2 - xmin_2) * (ymax_2 - ymin_2)  # [M, ]

    iou = overlaps / (area_1 + area_2 - overlaps)

    return iou


if __name__ == '__main__':
    import os
    os.environ["CUDA_VISIBLE_DEVICES"] = '13'
    boxes1 = np.array([[50, 50, 100, 300],
                       [60, 60, 100, 200]], np.float32)

    boxes2 = np.array([[50, 50, 100, 300],
                       [200, 200, 100, 200]], np.float32)

    print(iou_calculate_np(boxes1, boxes2))






# written by yjr

cimport cython
import numpy as np 
cimport numpy as np
import cv2
import time

DTYPE = np.float32
ctypedef np.float32_t DTYPE_t
ctypedef bint BOOL

cdef DTYPE_t two_boxes_iou(np.ndarray[DTYPE_t, ndim=1] rectangle_1, np.ndarray[DTYPE_t, ndim=1] rectangle_2):

    """
	calu rectangle_1 and rectangle_2 iou
    :param rectangle_1: [x, y, w, h, theta]. shape: (5, )
    :param rectangle_2:
    :return:
    """
    cdef DTYPE_t area1 = rectangle_1[2] * rectangle_1[3]
    cdef DTYPE_t area2 = rectangle_2[2] * rectangle_2[3]

    rect_1 = ((rectangle_1[0], rectangle_1[1]), (rectangle_1[3], rectangle_1[2]), rectangle_1[-1])
    rect_2 = ((rectangle_2[0], rectangle_2[1]), (rectangle_2[3], rectangle_2[2]), rectangle_2[-1])

    inter_points = cv2.rotatedRectangleIntersection(rect_1, rect_2)[1]

    cdef np.ndarray[DTYPE_t, ndim=3] order_points
    cdef float inter_area, iou
    if inter_points is not None:
        order_points = cv2.convexHull(inter_points, returnPoints=True)

        inter_area = cv2.contourArea(order_points)
        if area1 + area2 == inter_area:
            print ("area1-->", area1)
            print ("area2-->", area2)
            print ("inter_area-->", inter_area)
        iou = inter_area *1.0 / (area1 + area2 - inter_area)
        return <DTYPE_t> iou
    else:
        return <DTYPE_t> 0.0

cpdef np.ndarray[DTYPE_t, ndim=2] get_iou_matrix(
    np.ndarray[DTYPE_t, ndim=2] boxes1, # (N, 5)
    np.ndarray[DTYPE_t, ndim=2] boxes2): # (M, 5)
    
    cdef unsigned int num_of_boxes1 = boxes1.shape[0]
    cdef unsigned int num_of_boxes2 = boxes2.shape[0]

    cdef np.ndarray[DTYPE_t, ndim=2] iou_matrix = np.zeros((num_of_boxes1, num_of_boxes2), dtype=DTYPE)
    # cdef DTYPE_t box_iou 
    cdef unsigned int n, m
    # st = time.time()
    for n in range(num_of_boxes1):
        for m in range(num_of_boxes2):

            iou_matrix[n, m] = two_boxes_iou(boxes1[n], boxes2[m])
    # print "iou_matrix cost time: ", time.time() - st
    return iou_matrix




# --------------------------------------------------------
# Fast/er R-CNN
# Licensed under The MIT License [see LICENSE for details]
# Written by Bharath Hariharan
# --------------------------------------------------------
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import xml.etree.ElementTree as ET
import os
import pickle
import numpy as np

from utils import tools
from libs.label_name_dict.label_dict import LabelMap


class EVAL(object):
  def __init__(self, cfgs):
      self.cfgs = cfgs
      label_map = LabelMap(cfgs)
      self.name_label_map, self.label_name_map = label_map.name2label(), label_map.label2name()

  def write_voc_results_file(self, all_boxes, test_imgid_list, det_save_dir):
    '''

    :param all_boxes: is a list. each item reprensent the detections of a img.
    the detections is a array. shape is [-1, 6]. [category, score, xmin, ymin, xmax, ymax]
    Note that: if none detections in this img. that the detetions is : []

    :param test_imgid_list:
    :param det_save_path:
    :return:
    '''
    for cls, cls_id in self.name_label_map.items():
      if cls == 'back_ground':
        continue
      print("Writing {} VOC resutls file".format(cls))

      tools.makedirs(det_save_dir)
      det_save_path = os.path.join(det_save_dir, "det_"+cls+".txt")
      with open(det_save_path, 'wt') as f:
        for index, img_name in enumerate(test_imgid_list):
          this_img_detections = all_boxes[index]

          if this_img_detections.shape[0] == 0:
            continue

          this_cls_detections = this_img_detections[this_img_detections[:, 0]==cls_id]
          if this_cls_detections.shape[0] == 0:
            continue # this cls has none detections in this img
          for a_det in this_cls_detections:
            f.write('{:s} {:.3f} {:.1f} {:.1f} {:.1f} {:.1f}\n'.
                    format(img_name, a_det[1],
                           a_det[2], a_det[3],
                           a_det[4], a_det[5]))  # that is [img_name, score, xmin, ymin, xmax, ymax]

  def parse_rec(self, filename):
    """ Parse a PASCAL VOC xml file """
    tree = ET.parse(filename)
    objects = []
    for obj in tree.findall('object'):
      obj_struct = {}
      obj_struct['name'] = obj.find('name').text
      obj_struct['pose'] = obj.find('pose').text
      obj_struct['truncated'] = int(obj.find('truncated').text)
      obj_struct['difficult'] = int(obj.find('difficult').text)
      bbox = obj.find('bndbox')
      obj_struct['bbox'] = [int(bbox.find('xmin').text),
                            int(bbox.find('ymin').text),
                            int(bbox.find('xmax').text),
                            int(bbox.find('ymax').text)]
      objects.append(obj_struct)

    return objects

  def voc_ap(self, rec, prec, use_07_metric=False):
    """ ap = voc_ap(rec, prec, [use_07_metric])
    Compute VOC AP given precision and recall.
    If use_07_metric is true, uses the
    VOC 07 11 point method (default:False).
    """
    if use_07_metric:
      # 11 point metric
      ap = 0.
      for t in np.arange(0., 1.1, 0.1):
        if np.sum(rec >= t) == 0:
          p = 0
        else:
          p = np.max(prec[rec >= t])
        ap = ap + p / 11.
    else:
      # correct AP calculation
      # first append sentinel values at the end
      mrec = np.concatenate(([0.], rec, [1.]))
      mpre = np.concatenate(([0.], prec, [0.]))

      # compute the precision envelope
      for i in range(mpre.size - 1, 0, -1):
        mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

      # to calculate area under PR curve, look for points
      # where X axis (recall) changes value
      i = np.where(mrec[1:] != mrec[:-1])[0]

      # and sum (\Delta recall) * prec
      ap = np.sum((mrec[i + 1] - mrec[i]) * mpre[i + 1])
    return ap

  def voc_eval(self, detpath, annopath, test_imgid_list, cls_name, ovthresh=0.5,
                   use_07_metric=False, use_diff=False):
    '''

    :param detpath:
    :param annopath:
    :param test_imgid_list: it 's a list that contains the img_name of test_imgs
    :param cls_name:
    :param ovthresh:
    :param use_07_metric:
    :param use_diff:
    :return:
    '''
    # 1. parse xml to get gtboxes

    # read list of images
    imagenames = test_imgid_list

    recs = {}
    for i, imagename in enumerate(imagenames):
      recs[imagename] = self.parse_rec(os.path.join(annopath, imagename+'.xml'))
      # if i % 100 == 0:
      #   print('Reading annotation for {:d}/{:d}'.format(
      #     i + 1, len(imagenames)))

    # 2. get gtboxes for this class.
    class_recs = {}
    num_pos = 0
    # if cls_name == 'person':
    #   print ("aaa")
    for imagename in imagenames:
      R = [obj for obj in recs[imagename] if obj['name'] == cls_name]
      bbox = np.array([x['bbox'] for x in R])
      if use_diff:
        difficult = np.array([False for x in R]).astype(np.bool)
      else:
        difficult = np.array([x['difficult'] for x in R]).astype(np.bool)
      det = [False] * len(R)
      num_pos = num_pos + sum(~difficult)  # ignored the diffcult boxes
      class_recs[imagename] = {'bbox': bbox,
                               'difficult': difficult,
                               'det': det} # det means that gtboxes has already been detected

    # 3. read the detection file
    detfile = os.path.join(detpath, "det_"+cls_name+".txt")
    with open(detfile, 'r') as f:
      lines = f.readlines()

    # for a line. that is [img_name, confidence, xmin, ymin, xmax, ymax]
    splitlines = [x.strip().split(' ') for x in lines]  # a list that include a list
    image_ids = [x[0] for x in splitlines]  # img_id is img_name
    confidence = np.array([float(x[1]) for x in splitlines])
    BB = np.array([[float(z) for z in x[2:]] for x in splitlines])

    nd = len(image_ids) # num of detections. That, a line is a det_box.
    tp = np.zeros(nd)
    fp = np.zeros(nd)

    if BB.shape[0] > 0:
      # sort by confidence
      sorted_ind = np.argsort(-confidence)
      sorted_scores = np.sort(-confidence)
      BB = BB[sorted_ind, :]
      image_ids = [image_ids[x] for x in sorted_ind]  #reorder the img_name

      # go down dets and mark TPs and FPs
      for d in range(nd):
        R = class_recs[image_ids[d]]  # img_id is img_name
        bb = BB[d, :].astype(float)
        ovmax = -np.inf
        BBGT = R['bbox'].astype(float)

        if BBGT.size > 0:
          # compute overlaps
          # intersection
          ixmin = np.maximum(BBGT[:, 0], bb[0])
          iymin = np.maximum(BBGT[:, 1], bb[1])
          ixmax = np.minimum(BBGT[:, 2], bb[2])
          iymax = np.minimum(BBGT[:, 3], bb[3])
          iw = np.maximum(ixmax - ixmin + 1., 0.)
          ih = np.maximum(iymax - iymin + 1., 0.)
          inters = iw * ih

          # union
          uni = ((bb[2] - bb[0] + 1.) * (bb[3] - bb[1] + 1.) +
                 (BBGT[:, 2] - BBGT[:, 0] + 1.) *
                 (BBGT[:, 3] - BBGT[:, 1] + 1.) - inters)

          overlaps = inters / uni
          ovmax = np.max(overlaps)
          jmax = np.argmax(overlaps)

        if ovmax > ovthresh:
          if not R['difficult'][jmax]:
            if not R['det'][jmax]:
              tp[d] = 1.
              R['det'][jmax] = 1
            else:
              fp[d] = 1.
        else:
          fp[d] = 1.

    # 4. get recall, precison and AP
    fp = np.cumsum(fp)
    tp = np.cumsum(tp)
    rec = tp / float(num_pos)
    # avoid divide by zero in case the first detection matches a difficult
    # ground truth
    prec = tp / np.maximum(tp + fp, np.finfo(np.float64).eps)
    ap = self.voc_ap(rec, prec, use_07_metric)

    return rec, prec, ap

  def do_python_eval(self, test_imgid_list, test_annotation_path):
    AP_list = []
    # import matplotlib.pyplot as plt
    # import matplotlib.colors as colors
    # color_list = colors.cnames.keys()[::6]

    for cls, index in self.name_label_map.items():
      if cls == 'back_ground':
        continue
      recall, precision, AP = self.voc_eval(detpath=os.path.join(self.cfgs.EVALUATE_DIR,
                                                                 self.cfgs.VERSION),
                                            test_imgid_list=test_imgid_list,
                                            cls_name=cls,
                                            annopath=test_annotation_path,
                                            use_07_metric=self.cfgs.USE_07_METRIC,
                                            ovthresh=self.cfgs.EVAL_THRESHOLD)

      AP_list += [AP]
      print("cls : {}|| Recall: {} || Precison: {}|| AP: {}".format(cls, recall[-1], precision[-1], AP))
      # plt.plot(recall, precision, label=cls, color=color_list[index])
      # plt.legend(loc='upper right')
      # print(10*"__")
    # plt.show()
    # plt.savefig(cfgs.VERSION+'.jpg')
    print("mAP is : {}".format(np.mean(AP_list)))

  def voc_evaluate_detections(self, all_boxes, test_annotation_path, test_imgid_list):
    '''

    :param all_boxes: is a list. each item reprensent the detections of a img.

    The detections is a array. shape is [-1, 6]. [category, score, xmin, ymin, xmax, ymax]
    Note that: if none detections in this img. that the detetions is : []
    :return:
    '''
    test_imgid_list = [item.split('.')[0] for item in test_imgid_list]

    self.write_voc_results_file(all_boxes, test_imgid_list=test_imgid_list,
                           det_save_dir=os.path.join(self.cfgs.EVALUATE_DIR, self.cfgs.VERSION))
    self.do_python_eval(test_imgid_list, test_annotation_path=test_annotation_path)










# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf


def clip_boxes_to_img_boundaries(decode_boxes, img_shape):
    '''
    :param decode_boxes:
    :return: decode boxes, and already clip to boundaries
    '''

    with tf.name_scope('clip_boxes_to_img_boundaries'):

        # xmin, ymin, xmax, ymax = tf.unstack(decode_boxes, axis=1)
        xmin = decode_boxes[:, 0]
        ymin = decode_boxes[:, 1]
        xmax = decode_boxes[:, 2]
        ymax = decode_boxes[:, 3]
        img_h, img_w = img_shape[1], img_shape[2]

        img_h, img_w = tf.cast(img_h, tf.float32), tf.cast(img_w, tf.float32)

        xmin = tf.maximum(tf.minimum(xmin, img_w-1.), 0.)
        ymin = tf.maximum(tf.minimum(ymin, img_h-1.), 0.)

        xmax = tf.maximum(tf.minimum(xmax, img_w-1.), 0.)
        ymax = tf.maximum(tf.minimum(ymax, img_h-1.), 0.)

        return tf.transpose(tf.stack([xmin, ymin, xmax, ymax]))


def filter_outside_boxes(boxes, img_h, img_w):
    '''
    :param anchors:boxes with format [xmin, ymin, xmax, ymax]
    :param img_h: height of image
    :param img_w: width of image
    :return: indices of anchors that inside the image boundary
    '''

    with tf.name_scope('filter_outside_boxes'):
        xmin, ymin, xmax, ymax = tf.unstack(boxes, axis=1)

        xmin_index = tf.greater_equal(xmin, 0)
        ymin_index = tf.greater_equal(ymin, 0)
        xmax_index = tf.less_equal(xmax, tf.cast(img_w, tf.float32))
        ymax_index = tf.less_equal(ymax, tf.cast(img_h, tf.float32))

        indices = tf.transpose(tf.stack([xmin_index, ymin_index, xmax_index, ymax_index]))
        indices = tf.cast(indices, dtype=tf.int32)
        indices = tf.reduce_sum(indices, axis=1)
        indices = tf.where(tf.equal(indices, 4))
        # indices = tf.equal(indices, 4)
        return tf.reshape(indices, [-1])


def padd_boxes_with_zeros(boxes, scores, max_num_of_boxes):

    '''
    num of boxes less than max num of boxes, so it need to pad with zeros[0, 0, 0, 0]
    :param boxes:
    :param scores: [-1]
    :param max_num_of_boxes:
    :return:
    '''

    pad_num = tf.cast(max_num_of_boxes, tf.int32) - tf.shape(boxes)[0]

    zero_boxes = tf.zeros(shape=[pad_num, 4], dtype=boxes.dtype)
    zero_scores = tf.zeros(shape=[pad_num], dtype=scores.dtype)

    final_boxes = tf.concat([boxes, zero_boxes], axis=0)

    final_scores = tf.concat([scores, zero_scores], axis=0)

    return final_boxes, final_scores


# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import cv2
import tensorflow as tf
import sys

sys.path.append('../../')
from libs.utils.rotate_polygon_nms import rotate_gpu_nms
from libs.utils.coordinate_convert import coordinate5_2_8_tf


def nms_rotate(decode_boxes, scores, iou_threshold, max_output_size, use_gpu=True, gpu_id=0):
    """
    :param boxes: format [x_c, y_c, w, h, theta]
    :param scores: scores of boxes
    :param threshold: iou threshold (0.7 or 0.5)
    :param max_output_size: max number of output
    :return: the remaining index of boxes
    """

    if use_gpu:
        keep = nms_rotate_gpu(boxes_list=decode_boxes,
                              scores=scores,
                              iou_threshold=iou_threshold,
                              device_id=gpu_id,
                              max_output_size=max_output_size)

        keep = tf.cond(
            tf.greater(tf.shape(keep)[0], max_output_size),
            true_fn=lambda: tf.slice(keep, [0], [max_output_size]),
            false_fn=lambda: keep)

    else:
        keep = tf.py_func(nms_rotate_cpu,
                          inp=[decode_boxes, scores, iou_threshold, max_output_size],
                          Tout=tf.int64)
    return tf.cast(keep, tf.int64)


def nms_rotate_cpu(boxes, scores, iou_threshold, max_output_size):

    keep = []

    order = scores.argsort()[::-1]
    num = boxes.shape[0]

    suppressed = np.zeros((num), dtype=np.int)

    for _i in range(num):
        if len(keep) >= max_output_size:
            break

        i = order[_i]
        if suppressed[i] == 1:
            continue
        keep.append(i)
        r1 = ((boxes[i, 0], boxes[i, 1]), (boxes[i, 2], boxes[i, 3]), boxes[i, 4])
        area_r1 = boxes[i, 2] * boxes[i, 3]
        for _j in range(_i + 1, num):
            j = order[_j]
            if suppressed[i] == 1:
                continue

            if np.sqrt((boxes[i, 0] - boxes[j, 0])**2 + (boxes[i, 1] - boxes[j, 1])**2) > (boxes[i, 2] + boxes[j, 2] + boxes[i, 3] + boxes[j, 3]):
                inter = 0.0
            else:

                r2 = ((boxes[j, 0], boxes[j, 1]), (boxes[j, 2], boxes[j, 3]), boxes[j, 4])
                area_r2 = boxes[j, 2] * boxes[j, 3]
                inter = 0.0

                try:
                    int_pts = cv2.rotatedRectangleIntersection(r1, r2)[1]

                    if int_pts is not None:
                        order_pts = cv2.convexHull(int_pts, returnPoints=True)

                        int_area = cv2.contourArea(order_pts)

                        inter = int_area * 1.0 / (area_r1 + area_r2 - int_area + 1e-4)

                except:
                    """
                      cv2.error: /io/opencv/modules/imgproc/src/intersection.cpp:247:
                      error: (-215) intersection.size() <= 8 in function rotatedRectangleIntersection
                    """
                    # print(r1)
                    # print(r2)
                    inter = 0.9999

            if inter >= iou_threshold:
                suppressed[j] = 1

    return np.array(keep, np.int64)


def rnms_gpu(det_boxes, iou_threshold, device_id):
    if det_boxes.shape[0] == 0:
        return np.array([], np.int64)
    else:
        assert det_boxes.shape[1] == 6, 'shape of det_boxes is not 6, {}'.format(det_boxes)
        keep = rotate_gpu_nms(det_boxes, iou_threshold, device_id)
        keep = np.reshape(keep, [-1])
        return np.array(keep, np.int64)


def nms_rotate_gpu(boxes_list, scores, iou_threshold, device_id=0, max_output_size=100):
    # x_c, y_c, w, h, theta = tf.unstack(boxes_list, axis=1)
    # boxes_list = tf.transpose(tf.stack([x_c, y_c, w, h, theta]))

    # boxes_list_8 = coordinate5_2_8_tf(boxes_list)
    # xmin = tf.reduce_min(boxes_list_8[:, ::2], axis=1)
    # ymin = tf.reduce_min(boxes_list_8[:, 1::2], axis=1)
    # xmax = tf.reduce_max(boxes_list_8[:, ::2], axis=1)
    # ymax = tf.reduce_max(boxes_list_8[:, 1::2], axis=1)
    #
    # boxes_list_h = tf.transpose(tf.stack([xmin, ymin, xmax, ymax], axis=0))
    #
    # h_keep = tf.image.non_max_suppression(tf.reshape(boxes_list_h, [-1, 4]),  tf.reshape(scores, [-1, ]), max_output_size, iou_threshold)
    # boxes_list = tf.gather(boxes_list, h_keep)
    # scores = tf.gather(scores, h_keep)

    det_tensor = tf.concat([boxes_list, tf.expand_dims(scores, axis=1)], axis=1)
    # pad_box = tf.convert_to_tensor(np.array([[10, 10, 10, 10, -90, 0.001]], np.float32))
    # det_tensor = tf.concat([det_tensor, pad_box], axis=0)
    keep = tf.py_func(rnms_gpu,
                      inp=[det_tensor, iou_threshold, device_id],
                      Tout=tf.int64)
    keep = tf.reshape(keep, [-1])
    return keep


if __name__ == '__main__':
    boxes = np.array([[50, 50, 100, 100, 0],
                      [60, 60, 100, 100, 0],
                      [50, 50, 100, 100, -45.],
                      [200, 200, 100, 100, 0.]])
    boxes = np.tile(boxes, (10000, 1))

    scores = np.array([0.99, 0.88, 0.66, 0.77])
    scores = np.tile(scores, (10000))
    print(boxes.shape)


    keep = nms_rotate(tf.convert_to_tensor(boxes, dtype=tf.float32), tf.convert_to_tensor(scores, dtype=tf.float32),
                      0.7, 5)

    import os
    os.environ["CUDA_VISIBLE_DEVICES"] = '0'
    with tf.Session() as sess:
        print(sess.run(keep))
